/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S


#ifndef _ALCHEMY_H_
#define _ALCHEMY_H_

#include <stdio.h>
#include <X11/Xlib.h>
#include <picowidget.h>

#include "annotation_glue.h"

#ifdef DEBUG
#define DBG(x) x
#else
#define DBG(x)
#endif

#define MAX_PROPERTY_LENGTH 2048 // Maximum size of the window property
#define DEFAULT_THEME "/usr/local/share/alchemy/themes/pp/"
#define THEME_CONFIG_FILE "/usr/local/settings/alchemy/current_theme"
#define DEFAULT_FONT "-picopeta-simputer-bold-r-normal--11-120-101-108-p-120-fontspecific-0"
#define ELEVEN_POINT_FONT "-picopeta-simputer-bold-r-normal--11-120-101-108-p-120-fontspecific-0"
#define NINE_POINT_FONT "-picopeta-simputer-medium-r-normal--9-120-101-108-p-120-fontspecific-0"

// Special reserved APPIDs
#define AC_LAUNCHER_APPID	1
#define AC_STATUS_APPID		2
#define AC_KEYBOARD_APPID	3
#define AC_LOCK_APPID		11
#define AC_KEYBOARD_MENU_APPID	12
#define AC_CORONA_HELP_APPID	15
#define AC_IDENTITY_APPID	16
#define AC_XSTROKE_APPID	20
#define AC_CLOCK_APPID		25

// Commandlines to launch status, email, help and annotate
#define STATUS_CMDLINE "/usr/local/apps/status/status"
#define ANNOTATE_CMDLINE "/usr/local/apps/annotate/annotate %s %s grab"
#define ANNOTATE_NOGRAB_CMDLINE "/usr/local/apps/annotate/annotate %s %s nograb"
#define EMAIL_CMDLINE "/usr/local/apps/smile/smile %s \"%s\""
#define HELP_CMDLINE "/usr/local/apps/help_view/hlpview %s %s"
#define IDENTITY_CMDLINE "_AC_W_APPID=16 _NET_WM_NAME=Identity /usr/local/apps/help_view/hlpview /usr/local/store/annotations/identity/identity identity 10"
#ifdef __arm__
#define XCLIP_CMDLINE "/usr/local/bin/xclip"
#else
#define XCLIP_CMDLINE "/usr/bin/X11/xclip"
#endif
#define XSTROKE_LOCATION "/usr/local/apps/xstroke/xstroke"
#define XSTROKE_CMDLINE "/usr/local/apps/xstroke/xstroke -Ad -a /usr/local/apps/xstroke/alphabet -fn -picopeta-simputer-bold-r-normal--11-120-101-108-p-120-fontspecific-0 -g 30x16+%d+%d"

#ifdef __arm__
#define FLITE_CMDLINE "/usr/local/bin/flite"
#else
#define FLITE_CMDLINE "/usr/bin/flite"
#endif

#define HELP_TEMPLATE "/usr/local/store/help/%s/%s"
#define CHIKKI_HELP_TEMPLATE "/usr/local/store1/help/%s/%s"
#define BACKGROUND_GRAB_IMG "background.png"
#define ANNOTATE_HELP_KEY "__ANN_HELP__"
#define ANNOTATE_APP_NAME "Annotate"

// Pid file
#define PID_FILE "/tmp/alchemy.pid"

// The types of windows supported by the wm
typedef enum _window_type {
	NORMAL_WIN,			// Normal windows
	TOOLBAR_WIN,		// Toolbars like the keyboard
	DIALOG_WIN,			// Dialog windows
	SPLASH_WIN,			// Splash screens, no decorations
	MENU_WIN,			// Drop downs and popups.
} window_type;

// Areas (windows) where the event can occur
typedef enum _event_area {
	NOWHERE,
	TOPBAR_FRAME,
	BOTTOMBAR_FRAME,
	TOPLEVEL_FRAME,
	APP_FRAME
} event_area;

// The types of rotations supported (in degrees)
typedef enum _rotation_type {
	ROT_0,
	ROT_90,
	ROT_180,
	ROT_270,
} rotation_type;

// The list of ICCCM and EWMH atoms to be interpreted
// Please read ICCCM specs and EWMH specs on freedesktop.org
typedef enum _atomlist_index {
	WM_CLASS,
	WM_STATE,
	WM_SAVE_YOURSELF,
	WM_DELETE_WINDOW,
	WM_CHANGE_STATE,
	WM_PROTOCOLS,

	_NET_WM_STATE,
	_NET_WM_WINDOW_STATE_FULLSCREEN,

	_NET_WM_WINDOW_TYPE_TOOLBAR,	// A toolbar application sets this
	_NET_WM_WINDOW_TYPE_SPLASH,	// A splash screen sets this
	_NET_WM_WINDOW_TYPE_DIALOG,	// A transient(dialog) window sets this

	// Ewmh commands
	_NET_ACTIVE_WINDOW,		// Make me active window
	_NET_CLOSE_WINDOW,		// Close specified window

	// Ewmh info atoms
	_NET_SUPPORTED,	
	_NET_NUMBER_OF_DESKTOPS,	// Number of desktops this wm supports
	_NET_CURRENT_DESKTOP,		// Current desktop number
	_NET_CLIENT_LIST,		// List of windows
	_NET_SUPPORTING_WM_CHECK,
	
	_NET_WM_NAME,			// Application name: alls apps MUST set this
	_NET_WM_VISIBLE_NAME,		// Application title
	_NET_WM_WINDOW_TYPE,		// Type of the window; normal, toolbar or splash
	_NET_WM_PID,			// The PID of the application
	_NET_WM_PING,			// The WM Ping protocol
	_NET_WM_STATE_FULLSCREEN,	// Make me a full screen application

	_AC_WM_THEME,			// Theme used by alchemy
	_AC_WM_PID_MAP,			// Map of appids to pids
	_AC_WM_LAYER_MAP,		// Map of layer window ids to pids
	_AC_WM_LAUNCH_CMD,		// Appid, appname, cmdline triplet for launching

	_AC_WM_COMMAND,			// Message type for commands

	_AC_W_FLAGS,			// flags
	_AC_W_APPID,			// application's appid
	_AC_W_LAUNCHER,			// application's launcher's appid
	_AC_W_CMDLINE,			// Command line to launch program
	_AC_W_TITLE_FONT,		// Font for title
	_AC_W_TITLE_JUSTIFY,		// Justification of the title
	_AC_W_STATUS_MSG,		// Status message
	_AC_W_STATUS_FONT,		// Font for status message
	_AC_W_STATUS_JUSTIFY,		// Justfication of the status message
	_AC_W_PAGE_NUM,			// Page number
	_AC_W_PAGE_NUM_FONT,		// Font for page number
	_AC_W_PAGE_NUM_JUSTIFY,		// Justification of the page number
	_AC_W_KEYBOARD_LAYOUT,		// Keyboard Layout
	_AC_W_KEYBOARD_CATEGORY,	// Keyboard category
	_AC_W_ANNOTATE_KEY,		// Annotation key
	_AC_W_ANNOTATE_STATE,		// Annotation state (whether annotation exists)
	_AC_W_FAST_LOAD,		// Whether this is a fast load application
	_AC_W_CHIKKI,			// Whether this app is a chikki app
#ifdef __arm__
	_AC_W_ROTATION_STATE,		// Rotational state
#endif
	_AC_W_HELP_KEY,			// Help key
	_AC_W_EMAIL_FILE,		// File to be emailed
	_AC_W_DEL_ANNOTATION,		// Annotation key to be deleted
	_AC_W_LAYER_WINDOWID,		// Window ID to which I am the layer above

	_AC_BUTTON_UP,			// Up button pressed
	_AC_BUTTON_UP_REL,		// Up button released
	_AC_BUTTON_EMAIL,		// Email button pressed
	_AC_BUTTON_EMAIL_REL,		// Email button released
	_AC_BUTTON_NEXTPAGE,		// Nextpage button pressed
	_AC_BUTTON_NEXTPAGE_REL,	// Nextpage button released
	_AC_BUTTON_PREVPAGE,		// Prevpage button pressed
	_AC_BUTTON_PREVPAGE_REL,	// Prevpage button released
	_AC_BUTTON_PRINTER,		// Printer button pressed
	_AC_BUTTON_PRINTER_REL,		// Printer button released
	_AC_BUTTON_PREF,		// Preferences button pressed
	_AC_BUTTON_PREF_REL,		// Preferences button released
	_AC_BUTTON_RETRACTBAR,		// Retractable toolbar button pressed
	_AC_BUTTON_RETRACTBAR_REL,	// Retractable toolbar button released
	_AC_BUTTON_VOICE,		// Voice button pressed
	_AC_BUTTON_VOICE_REL,		// Voice button released
	_AC_BUTTON_SPEAKER,		// Speaker button pressed
	_AC_BUTTON_SPEAKER_REL,		// Speaker button released
	_AC_BUTTON_TICK,		// Tick button pressed
	_AC_BUTTON_TICK_REL,		// Tick button released
	_AC_BUTTON_CLOSE,		// Close button pressed (launcher only)
	_AC_BUTTON_CLOSE_REL,		// Close button released (launcher only)
	_AC_BUTTON_ROCKER_UP,		// Rocker up button pressed
	_AC_BUTTON_ROCKER_UP_REL,	// Rocker up button released
	_AC_BUTTON_ROCKER_DOWN,		// Rocker down button pressed
	_AC_BUTTON_ROCKER_DOWN_REL,	// Rocker down button released
	_AC_BUTTON_ROCKER_SELECT,	// Rocker select button pressed
	_AC_BUTTON_ROCKER_SELECT_REL,	// Rocker select button released
	_AC_ACCEL_RIGHT_FLICK,		// Accelerometer flicked right
	_AC_ACCEL_LEFT_FLICK,		// Accelerometer flicked left
	_AC_ACCEL_TOP_FLICK,		// Accelerometer flicked top
	_AC_ACCEL_BOTTOM_FLICK,		// Accelerometer flicked bottom
	_AC_ACCEL_RIGHT_EXTREME,	// Accelerometer tilted right
	_AC_ACCEL_LEFT_EXTREME,		// Accelerometer tilted left
	_AC_ACCEL_TOP_EXTREME,		// Accelerometer tilted top
	_AC_ACCEL_BOTTOM_EXTREME,	// Accelerometer tilted bottom

	XSTROKE_RECOGNITION_CMD,	//  For xstroke integration

	NR_ATOMS,
} atomlist_index;

// For xstroke integration
enum {
	XSTROKE_START_RECOGNITION=1,
	XSTROKE_STOP_RECOGNITION
};

// Index of flag bits
typedef enum _flag_num {
		CLOSE,
		UP,
		STATUS,
		PRINTER,
		SPEAKER,
		PREV,
		NEXT,
		HELP,
		PREF,
		LAUNCHER,
		ANNOTATE,
		ANNOTATION_PRESENT,
		EMAIL,
		INPUT,
		INPUT_CHOOSE,
		RETRACTBAR,
		ONETAP,
		PAGE_NAV,
		FULL_SCREEN,
		NEXT_PAGE_ENABLED,
		PREV_PAGE_ENABLED,
		STAR,
		LEFT,
		RIGHT,
		STATUS_MSG,
		NEEDS_KEYBOARD,
		NEEDS_PRINTER,
		NEEDS_ANNOTATION,
		VOICE,
		XSTROKE,
		NEEDS_INPUT,
		TICK,
		NR_BUTTONS
} flag_num;

// Status of various buttons to be shown
typedef union _window_flags {
	unsigned long all_flags;
	struct _bits {
		unsigned char close :1;		// Close an application
		unsigned char up :1;		// Up arrow button	
		unsigned char status :1;	// status launcher
		unsigned char printer :1;	// Printer icon
		unsigned char speaker :1;	// Speaker button
		unsigned char prev :1;		// previous application
		unsigned char next :1;		// next application
		unsigned char help :1;		// help button
		unsigned char pref :1;		// not implemened yet
		unsigned char launcher :1;	// home
		unsigned char annotate :1;	// Annotation - not yet there
		unsigned char annotation_present :1;
		unsigned char email :1;		// Send mail
		unsigned char input :1;		// Keyboard icon
		unsigned char input_choose :1;	// keyboard chooser
		unsigned char retractbar :1;    // retractable toolbar launcher
		unsigned char onetap :1;	// one tap launcher
		unsigned char page_nav :1;
		unsigned char full_screen :1;   /* not exactly a button*/
		unsigned char next_page_enabled :1;   /* previous page of an application*/
		unsigned char prev_page_enabled :1;   /* next page of an application*/
		unsigned char star :1;		// Launcher for mini-apps
		unsigned char left :1;		// Launcher left
		unsigned char right :1;		// Launcher right
		unsigned char status_msg :1;	// Status message visibility
		unsigned char needs_keyboard :1;
		unsigned char needs_printer :1;	// App wants to print
		unsigned char needs_annotation :1;
		unsigned char voice :1;		// For annotation
		unsigned char xstroke :1;	// For xstroke
		unsigned char needs_input :1;	// Needs either xstroke/keyboard
		unsigned char tick :1;		// Tick mark
	} bits;
} window_flags;

// Typedefs
typedef struct _wm wm;
typedef struct _window window;
typedef struct _ac_theme ac_theme;
typedef struct _winlist winlist;

#include "theme.h"
// Alchemy Theme
struct _ac_theme {
	Bool initialized;				// Whether theme has been initialized
	char *theme_path;				// Path of the theme file

	/*Pointer to the structure containing top level theme 
	 (topbar and bottombar) information*/
	theme_top_level_frame *top_level_frame;


	theme_menu *menu; 				// Theme info about star menu and title menu
	theme_pixmap *pixmap_tree;
	Window	 win; 					// required by pixmap creation functions
	wm *wm;	
	unsigned long menu_x; 			//x-pixel from where the menu will be displayed..will have the value 
									//from where the star button is placed.

	void (*load_theme)(ac_theme *self, const char *theme_path, Window root_win);	// Initialize the theme
	void (*render_theme)(ac_theme *self, window *win); // Allocate and draw toolbars
	void (*cleanup_theme)(ac_theme *self, window *win);// Deallocate and free data
	void (*free_theme)(ac_theme *self); //Free the theme structure

	/* Get various dimensions useful for stocking */

	unsigned long (*get_topbar_height)(ac_theme *self);
	unsigned long (*get_bottombar_height)(ac_theme *self);
	unsigned long (*get_dialogbar_height)(ac_theme *self);

	/* 
	 * If the value pointed to by bg_type_flag 
	 * is 0 this returns a pointer to px_pixmap; 
	 * else a pointer to px_color
	 */
	void *(*get_menu_background)(ac_theme *self, char *bg_type_flag);

	unsigned long (*get_menu_x)(ac_theme *self);	// Get menu position
};

// Structure for each window that is managed by Alchemy
struct _window {
	window *next,*prev;			// Doubly linked list of visible windows
	window *layer_above,*layer_below;	// Doubly linked list of hidden layers

	window *transient_of;		// I am a transient of

	window *transient_next;		// My transient windows
	window *transient_prev;		// My previous sibling of transients

	window_type wtype;		// The type of window
	wm *wm;				// Pointer to window manager

	px_widget **frame_widgets;	// Quick access to various button/label widgets

	unsigned short x;		// Origin top left corner
	unsigned short y;
	unsigned short height;		// Dimensions
	unsigned short width;
	
	unsigned long appid;		// application UID
	unsigned long launcher_appid;	// launched by this app
	char *name;			// application's unique name ??

	char *title;			// title visible UTF8 text
	unsigned char tjustify;		// title justification
	char *tfont;			// Font for title

	char *status_msg;		// status message
	unsigned char sjustify;		// Status justification
	char *sfont;			// Font for title

	char *page_num;			// Page navigation number
	unsigned char pjustify;		// Page number justification
	char *pfont;			// Font for page number
	
	Window frame_window;		// Top level container window
	Window app_window;		// Application window
	px_widget *topbar_frame;
	px_widget *bottombar_frame;	// Bottom bar frame
	Bool ignore_next_unmap;		// Ignore the next unmap event on this window
	Bool onetap;			// The next mail to be sent is a onetap mail
	Bool app_speaks;		// This application handles speaker events
	Bool is_mapped;			// Whether this window is mapped or not
	Bool fast_load;			// Whether this window must be fast-loaded
	Bool fast_visible;		// Whether this fast-loaded window is currently visible
	Bool needs_grab;		// Whether the background should be grabbed for annotate
	Bool is_chikki;			// Whether the app is a chikki

	window_flags flags;		// All the button's visibilities and states
	window_flags oflags;		// All the button's previous state

	pid_t pid;			// pid of the process (if non zero)

	char *annotation_key;		// Annotation key
	char *help_key;			// Help key
	char *keyboard_layout;		// Keyboard layout
	char *keyboard_category;	// Keyboard category
#ifdef __arm__
	rotation_type rotation;		// The display orientation of this client
#endif

	void (*redraw)(window *self);
	void (*refresh_widgets)(window *self); // For dynamic theme changing
	void (*set_size)(window *self, unsigned long width, unsigned long height);
	void (*set_position)(window *self, int x, int y);
	void (*get_size)(window *self, unsigned long *width, unsigned long *height);
	void (*get_position)(window *self, int *x, int *y);
	void (*show)(window *self);
	void (*hide)(window *self);
	void (*iconify)(window *self);
	void (*destroy)(window *self);

	void (*set_flags)(window *self, unsigned long flags);
	void (*set_flag)(window *self, flag_num flag_no, Bool value);
};

// A List for managing alternate window lists not
// part of the main window structure, such as on_top list
struct _winlist {
	winlist *next;		// Next window in the winlist
	window *win;		// Pointer to window
};

// The singleton window manager structure
struct _wm {
	window *window_list;		// List of Windows, head of circular list
	window *mapped_window;		// Currently shown normal window
	window *focused_window;		// Currently focused window
	window *launcher_window;	// Pointer to launcher window
	window *keyboard_window;	// Pointer to input window
	window *status_window;		// Pointer to status window
	window *xstroke_window;		// Pointer to xstroke window
	window *clock_window;		// Pointer to clock window
	window *identity_save_window;	// Pointer to saved identity window

	Bool is_keyboard_mapped;	// Whether keyboard window is mapped
	Bool is_menu_mapped;		// Whether any menu window is mapped
	Bool is_printer_availible;	// Whether a printer is availible

	winlist *on_top_list;		// List of windows always on top

	Display *display;		// Display
	int screen;			// Screen
	
	Window root;			// Display root
	unsigned long black_pixel;	// Default blk/white pixels
	unsigned long white_pixel;	

	Atom atomlist[NR_ATOMS];

	ac_theme *current_theme;	// Current theme
	char *current_kbd_category;	// Current  keyboard category
	char *current_kbd_layout;	// Current  keyboard layout

	unsigned long display_width, display_height; // Display dimensions

	unsigned long num_windows;	// Number of windows managed by this wm
	Cursor normal_cursor;		// The default window cursor

#ifdef __arm__
	rotation_type current_rot;	// Current Rotation orientation
	Bool (*rotate_to)(wm *,rotation_type rot);	// Rotates display to specification
#endif
	void (*mainloop)(wm *);
	void (*exit)(wm *);
	void (*handle_event)(wm *, XEvent *);
	unsigned int (*get_toolbar_displacement)(wm *);
	void (*retile)(wm *, window *, Bool);
};

wm  *ac_init();
void ac_mainloop(wm *wm);
void ac_exit(wm *wm);
void ac_handle_event(wm *wm, XEvent *xev);
unsigned int ac_get_toolbar_displacement(wm *wm);
void ac_retile(wm *wm, window *w, Bool got_mapped);
window *ac_get_window(wm *wm, Window win);
void ac_make_pid_appid_map(wm *wm);
ac_theme *new_theme(wm *wm);

#define For_All_Mappable_PS_Do(wm,ps,twl) { \
			(twl)=(ps); \
			if((twl)) \
			do { \
			
#define For_All_Mappable_PS_Done(wm,ps,twl) (twl)=(twl)->next; \
			} while((twl) != (ps)); }

#define For_All_Windows_Do(wm,twl) { \
			window *twc; \
			twc=wm->window_list; \
			if(twc) \
			do { \
				for((twl)=twc;(twl);(twl)=(twl)->layer_below) { \
			
#define For_All_Windows_Done(wm,twl) }\
			  twc=twc->next; \
			} while(twc != wm->window_list); }

#define For_All_Mappable_Do(wm,twl) { \
			(twl)=wm->window_list; \
			if((twl)) \
			do { \
			
#define For_All_Mappable_Done(wm,twl) (twl)=(twl)->next; \
			} while((twl) != wm->window_list); }

#endif

#define mystrlen(data) data ? strlen(data) : 0

#define TMP_BUF_SIZE 256
#define ONE_TAP_STRIP_HEIGHT 25
